import json
from select import *

### Schema ###
### This class is used to load and store a dictionary from a config file.
### It contains methods to load data, dump data, and retrieve data from the dictionary.

class Schema():

    # Initialisation 
    def __init__(self):
        self.data = None
        self.internal = None


    # Gets a list by its dot-notated key.
    def get_select(self, k):
        if k in self.internal['selects'].keys():
            return Select(self.internal['selects'][k])
        return None


    # Gets a label by its dot-notated key.
    def get_label(self, label):
        if label in self.internal['labels'].keys():
            return self.internal['labels'][label]
        return label


    # Adds 'val' to 'key' in dot notation.
    def get_key_string(self, key, val):
        if key:
            return '.'.join([key, val])
        return val


    #Converts the loaded dictionary for use with the Fabric API.
    def convert(self, input):
        if isinstance(input, dict):
            return {self.convert(key): self.convert(value) for key, value in input.iteritems()}
        elif isinstance(input, list):
            return [self.convert(element) for element in input]
        elif isinstance(input, unicode):
            return input.encode('utf-8')
        else:
            return input


    # Gets a value from dot format: s.get('project.cache.backend') from dict format: self.data['project']['cache']['backend']
    def get(self, key):
        if not self.data:
            return None

        # Split the key by its dots.        
        parts = key.rsplit('.')

        # Loop through the key.
        v = self.data
        for p in parts:
            
            # Try and get the value for this section in the key.
            try:
                v = v.get(p)

            # If the key is invalid, return None rather than throwing an exception.
            except AttributeError as e:
                return None

        # If we got the value successfully, return it.
        return v


    # Loads self.data from a file.
    def load(self, path):
        data = None

        # Try and open the file, or raise an exception.
        try:
            handle = open(path)
        except IOError as e:
            raise e

        # Try to load valid JSON, or raise an exception.
        try:
            data = json.load(handle)
        except ValueError as e:
            raise e

        # Ensure that the JSON is at least a dictionary, or raise an exception.
        if type(data) is not dict:                    
            raise Exception("Invalid schema")

        # Store the dictionary.
        handle.close()
        self.data = self.convert(data)

        if '_internal' in self.data.keys():
            self.internal = self.data['_internal']
            del(self.data['_internal'])

        return self

    # Saves self.data to a file.
    def save(self, path, as_schema = False):

        if as_schema:
            self.data['_internal'] = self.internal

        # Try and save the file to the path.
        try:
            handle = open(path, 'w')
            json.dump(self.data, handle, indent = 4)

        # Raise an exception if we can't find the file.
        except IOError as e:
            raise e
        finally:
            if '_internal' in self.data.keys():
                del(self.data['_internal'])

        handle.close()


import argparse
import lazyconf

from lib.schema import *
from lib.prompt import *
from lib.select import *
from lib.merge import *

# A function to be called from the setup entrypoint.
def parse():
    parser = argparse.ArgumentParser(prog='lazyconf')
    sp = parser.add_subparsers(dest='command')

    sp_config = sp.add_parser('config', help='Run the configurator.')
    
    sp_get = sp.add_parser('get', help='Gets the value for a specific key in the configuration.')
    sp_get.add_argument('-k', '--key',
                        help='The key to retrieve.', dest='key')

    return parser.parse_args()

def console():
    command_dict = parse().__dict__
    command = command_dict['command']

    l = lazyconf.Lazyconf()

    if command == 'get':
        try:
            l.load()
        except:
            exit()
        r = l.get(command_dict['key'])
        if r:
            print r
        exit()

    elif command == 'config':
        l.configure()
       
# Call the console function if we call this file directly.
if __name__ == '__main__':
    console()



import os 
import unittest
import lazyconf

from lib.schema import *
from lib.prompt import *
from lib.select import *
from lib.merge import *

class LazyTests(unittest.TestCase): 

    # Individually test the Merge class.
    def test_merge(self):
        schema = {
            "present1" : True,
            "present2" : "foo",
            "present3" : 12,
            "present4" : {
                "p1" : 1,
                "p2" : "too",
                "p3" : True
            },

            "added1" : False,
            "added2" : "bar",
            "added3" : 24,
            "added4" : [1,2,3,4,5],


            "modified1" : {
                "a" : 1,
                "b" : True,
                "c" : "baz"
            },

            "modified2" : [5,4,3,2,1],
            "modified3" : [24,45,23,87,4],
        }

        data = {
            "present1" : False,
            "present2" : "",
            "present3" : 0,
            "present4" : {
                "p1" : 0,
                "p2" : "",
                "p3" : False
            },

            "modified1" : "foobar",
            "modified2" : 12345,
            "modified3" : True,
            "modified4" : True,

            "removed1" : "remo",
            "removed2" : 23,
            "removed3" : {
                "all" : "of",
                "us" : "will",
                "be" : "removed"
            }
        }

        merge = Merge(schema, data)
        mods = merge.merge()

        return


    # Individually test the Schema class.
    def test_schema(self):
        p = os.path.dirname(lazyconf.__file__)
        if not p:
            raise Exception("Could not find local path.")

        s = Schema()

        val = s.get('db.engine')

        with self.assertRaises(ValueError):
            s.load(p + '/schema/test/invalid.json')

        with self.assertRaises(IOError):
            s.load(p + '/schema/test/noexist.json')

        with self.assertRaises(Exception):
            s.load(p + '/schema/test/noobject.json')

        s.load(p + '/schema/test/valid.json')

        select = s.get_select('db.engine')
        select = s.get_select('db.noexist')
        label = s.get_label('db.engine')
        label = s.get_label('db.noexist')

        key1 = s.get_key_string('','one')
        key2 = s.get_key_string('one','two')

        val = s.get('db.engine')
        val = s.get('db.engine.noexist')

        s.save('/dev/null')

        with self.assertRaises(IOError):
            s.save('/noexist/does_not_exist')

        return


    # Individually test the Prompt class.
    def test_prompt(self):
        p = Prompt()

        p.header("Testing.")
        p.success("Testing.")
        p.error("Testing.")
        p.notice("Testing.")

        return


    # Individually test the Select class.
    def test_select(self):
        return

if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(LazyTests)
    unittest.TextTestRunner(verbosity=2).run(suite)



### Select ###
### This class contains functions for getting keys from values, values from keys, and regex helpers for selecting a value from a list.

class Select():

    # Initialisation.
    def __init__(self, d):
        self.dict = d


    # Gets the key from a value in a select.
    def get_key(self, v):
        for k,val in self.dict.iteritems():
            if v == val:
                return k
        return ""


    # Gets the value from a key in a select.
    def get_value(self, k):
        if k in self.dict.keys():
            return self.dict[k]
        return ""

    def first_choice(self):
        return sorted(self.dict.keys())[0]

    def choices(self):
        return '(' + ','.join(sorted(self.dict.keys())) + ')'


    def reg_choices(self):
        return '^(' + '|'.join(self.dict.keys()) + ')$'.encode('string_escape');


import os
import lazyconf

from lib.schema import *
from lib.prompt import *
from lib.select import *
from lib.merge import *

### Lazyconf ###
### Our main class. Annotate public functions better.

class Lazyconf():

    lazy_folder = '.lazy/'
    ignore_file = '.gitignore'
    schema_filename = 'lazy.schema.json'
    data_filename = 'lazy.json'

    # Initialisation.
    def __init__(self):
        self.prompt = Prompt()
        self.data = None
    

    # Creates the self.lazy_folder if it doesn't already exist, and writes the .gitignore.
    def add_ignore(self, line):
        path = self.lazy_folder + self.ignore_file
        if os.path.isfile(os.path.realpath(path)):
            self.prompt.notice("OKIE DOKE")
            return

        try:
            handle = open(path,'w')
            handle.write(line + '\n')
        except IOError as e:
            raise e

        handle.close()


    # Finds all schema templates and prompts to choose one. Copies the file to self.lazy_folder.
    def choose_schema(self, out_file):

        path = os.path.dirname(lazyconf.__file__) + '/schema/'
        self.prompt.header('Choose a template for your config file: ')

        i = 0
        choices = []

        for filename in os.listdir(path):
            if filename.endswith('.json'):
                try:
                    template = self._load(path + filename)
                    description = template.get('_meta.description')
                    prompt_string = str(i + 1) + '. ' + filename
                    i += 1

                    if description:
                        self.prompt.notice(prompt_string + ': ' + description)
                    else:
                        self.prompt.notice(prompt_string)

                    choices.append(template)

                except IOError as e:
                    print self.prompt.error(str(e))

        val = 0
        while val is 0 or val > i:
            val = self.prompt.int('Choice', default = 1)
            if val is 0 or val > i:
                self.prompt.error('Please choose a value between 1 and ' + str(i) + '.')
        
        schema = choices[val-1]
        
        if '_meta' in schema.data.keys():
            del(schema.data['_meta'])
        
        schema.save(out_file, as_schema=True)
        sp, sf = os.path.split(out_file)
        self.prompt.success('Saved to ' + self.lazy_folder + sf + '.')
        return schema


    # Goes through all the options in the data file, and prompts new values.
    def configure_data(self, data, key_string = ''):
        
        # If there's no keys in this dictionary, we have nothing to do.
        if len(data.keys()) == 0:
            return

        # Split the key string by its dots to find out how deep we are.
        key_parts = key_string.rsplit('.')
        prefix = '  ' * (len(key_parts) - 1)

        # Attempt to get a label for this key string.
        label = self.data.get_label(key_string)

        # If we are have any key string or label, write the header for this section.
        if label:
            p = prefix
            if len(p) > 0:
                p += ' '
            self.prompt.header(p + '[' + label + ']')

        # Add to the prefix to indicate options on this level.
        prefix = prefix + '   '

        # If this section has an '_enabled' key, process it first, as it could enable or disable this whole section.
        if '_enabled' in data.keys():
            s = self.data.get_key_string(key_string, '_enabled')

            #Prompt whether to enable this section. Use the existing value as the default.
            data['_enabled'] = self.prompt.bool(prefix + self.data.get_label(s), data['_enabled'])

            # Return if this section is now disabled.
            if data['_enabled'] is False:
                return

        # Loop through the rest of the dictionary and prompt for every key. If the value is a dictionary, call this function again for the next level.
        for k, v in data.iteritems():

            # If we hit the '_enabled' key, we've already processed it (but still need it in the dictionary for saving). Ignore it.
            if k == '_enabled':
                continue
            
            # Get the type of the value at this key, and the dot-noted format of this key.
            t = type(v)
            s = self.data.get_key_string(key_string, k)

            # See if this key has an associated select. If so, this type is a select type.
            select = self.data.get_select(s)
            if select:
                data[k] = self.prompt.select(prefix + self.data.get_label(s), select, default = v)

            # If the value type is a dictionary, recall this function.
            elif t is dict:
                self.configure_data(v, s)

            # If the value type is a boolean, prompt a boolean.
            elif t is bool:
                data[k] = self.prompt.bool(prefix + self.data.get_label(s), default = v)

            # If the value is an int, prompt and int.
            elif t is int:
                data[k] = self.prompt.int(prefix + self.data.get_label(s), default = v)

            # If someone has put a list in data, we default it to an empty string. If it had come from the schema, it would already be a string.
            elif t is list:
                data[k] = self.prompt.string(prefix + self.data.get_label(s) + ':', default = '')

            # If none of the above are true, it's a string.
            else:
                data[k] = self.prompt.string(prefix + self.data.get_label(s) + ':', default = v)


    # Loads the schema from a schema file.
    def configure(self):

        path = os.getcwd() + '/' + self.lazy_folder
        if not os.path.exists(path):
            os.makedirs(path)

        schema_file = path + self.schema_filename
        data_file = path + self.data_filename
        out_file = data_file

        # Initialise the schema and data objects.
        schema, data = Schema(), Schema()
        
        # Load the schema from a file.
        try:
            schema.load(schema_file)
        except (Exception, IOError, ValueError) as e:

            # If we can't load the schema, choose from templates.
            schema = self.choose_schema(schema_file)
        else:
            sp, sf = os.path.split(schema_file)
            self.prompt.success('Loaded schema from ' + self.lazy_folder + sf)

        # Load the data from a file.
        try:
            data.load(data_file)
        except (Exception, IOError, ValueError) as e:
            self.prompt.error('Could not find data file. Copying from schema...')
        else:
            sp, sf = os.path.split(data_file)
            self.prompt.success('Loaded data from ' + self.lazy_folder + sf)

        # Store the internals of the schema (labels, selects, etc.) in data.
        data.internal = schema.internal

        # If we have data from a data file, merge the schema file into it.
        if data.data:

            # Create a new Merge instance using the data from the schema and data files.
            m = Merge(schema.data, data.data)
            mods = m.merge()

            for a in mods['added']:
                self.prompt.success('Added ' + a + ' to data.')

            for r in mods['removed']:
                self.prompt.error('Removed ' + r + ' from data.')

            for k,m in mods['modified']:
                self.prompt.notice('Modified ' + k + ': ' + m[0] + ' became ' + m[1] + '.' )

        # Otherwise, reference the data from the schema file verbatim.
        else:
            data.data = schema.data

        # Store the data.
        self.data = data

        # Configure the data.
        self.configure_data(data.data)

        # Save the data to the out file.
        self.data.save(out_file)

        sp, sf = os.path.split(out_file)
        self.add_ignore(sf)
        self.prompt.success('Saved to ' + self.lazy_folder + sf + '.')


    # Get the data for a dot-notated key.
    def get(self, key):
        return self.data.get(key)


    def _load(self, data_file):

       # Load the data from a file.
        try:
            data = Schema().load(data_file)
        except (Exception, IOError, ValueError) as e:
            raise e

        return data

    # Takes a folder and loads the data from .lazy/ in that folder.
    def load(self, data_file = None):
        if not data_file:
            data_file = ''
        elif data_file[-1] != '/':
            data_file += '/'

        if data_file[-6:] != self.lazy_folder:
            data_file += self.lazy_folder
        
        data_file += self.data_filename

        self.data = self._load(data_file)
        return self



### Merge ###
### This class is used to merge variables in the schema dictionary into the data dictionary, without losing defaults.
### It never returns the dictionaries, as the original dictionaries are passed by reference automatically.
### The schema dictionary is never be modified by this class.

class Merge():

    # Initialisation creates sets of the top level keys in both dictionaries, and the intersection between the two.
    def __init__(self, schema, data):
        self.schema = schema
        self.data = data

        # Create the sets.
        self.set_schema = set(self.schema.keys())
        self.set_data = set(self.data.keys())
        self.intersect = self.set_schema.intersection(self.set_data)
        

    # Returns all top level keys which have been added to the schema.
    def added(self):
        return self.set_schema - self.intersect


    # Returns all top level keys which have been removed from the schema.
    def removed(self):
        return self.set_data - self.intersect


    # This method recursively merges all changes in the schema into the data file.
    def merge(self, schema = None, data = None, prefix = None):
        
        # If there's no prefix yet, don't add a '.'
        p = ''
        if prefix:
            p = prefix + '.'
        
        # Create the dictionary to be returned.
        mods = {
            'added'   : [], # []string
            'removed' : [], # []string
            'modified': []  # [](string,(type,type))
        }

        if schema is None:
            schema = self.schema

        if data is None:
            data = self.data

        # Add all new keys in schema to data.
        added = self.added()
        if added:
            for a in added:
                mods['added'].append(p + a)
                data[a] = schema[a]

        # Remove all keys in data that are no longer in schema.
        removed = self.removed()
        if removed:
            for r in removed:
                mods['removed'].append(p + r)
                del(data[r])

        # Recursively call this method until all common variables are resolved.
        intersect = self.intersect
        if intersect:
            for i in intersect:
                s = type(schema[i])
                d = type(data[i])

                # If there is a dictionary that is shared at this key by both the schema and the data, call this method again.
                if s is dict and d is dict:
                    
                    # Create a new instance of this class with the child dicts as schema and data.
                    diff = Merge(schema[i], data[i])
                    
                    # Merge the child dicts.
                    add_mods = diff.merge(prefix = p + i)

                    # Update our return dict.
                    mods['added'] += add_mods['added']
                    mods['removed'] += add_mods['removed']
                    mods['modified'] += add_mods['modified']
                    
                    # Remove the created Merge.
                    del(diff)

                # Otherwise if the variable type has changed, copy the variable from the schema to the data.
                elif s != d:
                    
                    # Here, we process lists into empty strings.
                    if s is list:
                            data[i] = ""

                    # If we're not dealing with a list, process normally.
                    else: 
                        # Add this to the 'modified' list.
                        m = (p + i, (str(d),str(s)))
                        mods['modified'].append(m)

                        # Copy the value from the schema to the data.
                        data[i] = schema[i]

        # Return the dictionary of modifications.
        return mods


import sys
from setuptools import setup, find_packages
import lazyconf

l = 'lazyconf/'
s = 'schema/'

setup(
    name='lazyconf',
    version='0.4.3',
    author='Fareed Dudhia',
    author_email='fareeddudhia@gmail.com',
    packages=find_packages(),
    package_data={"lazyconf": ['schema/' 'lazyconf/schema/*.json',]},
    include_package_data=True,
    py_modules=['lazyconf.lazyconf','lazyconf.console'],
    entry_points={
        'console_scripts': ['lazyconf = lazyconf.console:console',]},
    url='https://www.github.com/fmd/lazyconf',
    license='LICENSE.rst',
    description='Insultingly simple configuration for Python 2.7 applications.',
    long_description=open('README.rst').read(),
)



import re
from colors import Colors
### Prompt ###
### This class contains several helper functions for getting data to and from the end-user's input and a schema.Schema object.

class Prompt():

    # Initialisation
    def __init__(self):
        pass

    ### Formatting ###

    # Prints a header around a string.
    def header(self, msg):
        print(Colors.header + msg + Colors.end)


    # Prints a success message.
    def success(self, msg):
        print(Colors.success + msg + Colors.end)


    # Prints an error message.
    def error(self, msg):
        print(Colors.error + msg + Colors.end)


    # Prints a notice message.
    def notice(self, msg):
        print(Colors.blue + msg + Colors.end)


    # The prompt string method.
    # This method heavily models how it works in Fabric.
    def string(self, label, default='', validate=None):
        
        p = label + ' ' + '[%s]' % str(default).strip() + ' '
        value = None
        while value is None:

            # Get the raw input.
            value = raw_input(p) or default
            if validate:

                # Attempt to validate the string if it's callable.
                if callable(validate):
                    try:
                        value = validate(value)
                    except Exception, e:

                        # If the regex fails, loop.
                        value = None
                        self.error("Validation failed for the following reason:")
                        self.error(indent(e.message) + '\n')

                # Attempt to validate it if it's not callable.
                else:
                    if not validate.startswith('^'):
                        validate = r'^' + validate

                    if not validate.endswith('$'):
                        validate += r'$'

                    result = re.findall(validate, value)
                    if not result:
                        self.error("Regular expression validation failed: '%s' does not match '%s'\n" % (value, validate))

                        # If the regex fails, loop.
                        value = None

        return value

    ### Prompts ###

    # Returns the value from a validated int prompt.
    def int(self, label, default = 0):
        val = self.string(label + ' (int):', default = str(default), validate = r'^[0-9]+$')
        return int(val)


    # Returns the value from a validated bool prompt.
    def bool(self, label, default = False):
        val = self.string(label + ' (y/n):', default = self.fmt_bool(default), validate = r'^(y|n)$')
        return self.defmt_bool(val)


    # Returns the value from a validated select prompt.
    def select(self, label, select, default = ""):
        deft = select.get_key(default)
        if not deft:
            deft = select.first_choice()
        return select.get_value(self.string(label + ' ' + select.choices() + ':', default = deft, validate = select.reg_choices()))


    ### Formatting ###

    # Takes a python bool and returns it as y|n.
    def fmt_bool(self, p):
        if p is True:
            return 'y'
        if p is False:
            return 'n'
        return ''


    # Takes y|n and returns a python bool.
    def defmt_bool(self, p): 
        if p == 'y':
            return True
        if p == 'n':
            return False
        return None